#!/usr/bin/env python3

"""Converts adt_YYYY-MM-DD.mat files to two collections of shapefiles,
SHPC_anti and SHPC_cyclo. The mat files are assumed to be in v7.3
(HDF5) format and will be converted to v6 format before being read in
Python. The data in each input file is the set of detected
instantaneous eddies at a given date. The second, optional argument,
is the final date processed.

"""

import shapefile
import numpy as np
from os import path
import scipy.io as sio
import datetime
import f90nml
import sys
import os
from dateutil import parser
import subprocess

def define_fields(writers):
    """writers is a dictionary of shapefile.Writer objects."""
    
    writers["extr"].field("ssh", "N", 13, 6)
    writers["extr"].field("days_1950", "N", 5)
    writers["extr"].field("eddy_index", "N", 5)
    writers["extr"].field("valid", "N", 1)
    writers["extr"].field("speed", "N", 13, 6)

    writers["centroid"].field("days_1950", "N", 5)
    writers["centroid"].field("eddy_index", "N", 5)

    writers["outer"].field("r_eq_area", "N", 10, 4)
    writers["outer"].field("ssh", "N", 13, 6)
    writers["outer"].field("days_1950", "N", 5)
    writers["outer"].field("eddy_index", "N", 5)

    writers["max_speed"].field("r_eq_area", "N", 10, 4)
    writers["max_speed"].field("ssh", "N", 13, 6)
    writers["max_speed"].field("days_1950", "N", 5)
    writers["max_speed"].field("eddy_index", "N", 5)

def write(writers, cell, cyclone):
    """Write to shapefiles eddies with a given orientation, at a given
    date.

    writers is a dictionary. cell is a numpy.ndarray with shape
    (number of eddies, 21). cyclone is a boolean. The eddy
    identification numbers start at 1 and increment by 1 for each eddy
    in cell.

    """
    eddy_index = 0

    for eddy in cell:
        # eddy is a numpy.ndarray with shape (21,). The
        # description of each element of eddy is given by
        # matlab_data["Fields"]. Note that amplitudes in
        # eddy[7] and eddy[11] are positive regardless of eddy
        # orientation.
        eddy_index += 1

        writers["extr"].point(eddy[0], eddy[1])
        i = np.argwhere(matlab_data["X"] == eddy[0]).item()
        j = np.argwhere(matlab_data["Y"] == eddy[1]).item()
        speed = 1e4 if np.isnan(eddy[12]) else eddy[12]
        writers["extr"].record(ssh = matlab_data["ADT"][i, j], valid = 1,
                               days_1950 = d, eddy_index = eddy_index,
                               speed = speed)

        if cyclone:
            ssh = matlab_data["ADT"][i, j] + eddy[7]
        else:
            ssh =  matlab_data["ADT"][i, j] - eddy[7]

        writers["centroid"].point(eddy[2], eddy[3])
        writers["centroid"].record(days_1950 = d, eddy_index = eddy_index)

        writers["outer"].record(r_eq_area = eddy[6], ssh = ssh, days_1950 = d,
                                eddy_index = eddy_index)
        polyline = np.stack((eddy[4], eddy[5]), axis = 1).tolist()
        writers["outer"].poly([polyline])

        if cyclone:
            ssh = matlab_data["ADT"][i, j] + eddy[11]
        else:
            ssh =  matlab_data["ADT"][i, j] - eddy[11]

        writers["max_speed"].record(r_eq_area = eddy[10], ssh = ssh,
                                    days_1950 = d, eddy_index = eddy_index)
        polyline = np.stack((eddy[8], eddy[9]), axis = 1).tolist()
        writers["max_speed"].poly([polyline])

    n_shapes = len(writers["extr"])
    writers["ishape_last"].write(str(n_shapes - 1) + "\n")

if len(sys.argv) == 1: sys.exit("Required argument: first input .mat file")
adt_file = sys.argv[1]
dirname, basename = path.split(adt_file)
valid_date_formats = ["adt_%Y-%m-%d.mat", "adt_%Y%m%d.mat"]

for date_format in valid_date_formats:
    try:
        my_date = datetime.datetime.strptime(basename, date_format).date()
    except ValueError:
        pass
    else:
        break

d = my_date - datetime.date(1950, 1, 1)
d = d.days

if len(sys.argv) == 3:
    final_date = parser.parse(sys.argv[2], yearfirst=True, dayfirst = False)\
                       .date()
else:
    final_date = my_date

# There are two factories, one for cyclones and one for anticyclones:
factories = [dict(writers = {}, cell_name = "Anticyclonic_Cell",
                  cyclone = False, SHPC = "SHPC_anti"),
             dict(writers = {}, cell_name = "Cyclonic_Cell", cyclone = True,
                  SHPC = "SHPC_cyclo")]

for factory in factories:
    fname = path.join(factory["SHPC"], "extremum")
    factory["writers"]["extr"] = shapefile.Writer(fname,
                                                  shapeType = shapefile.POINT)
    
    fname = path.join(factory["SHPC"], "centroid")
    factory["writers"]["centroid"] = shapefile.Writer(fname,
                                                shapeType = shapefile.POINT)
    
    fname = path.join(factory["SHPC"], "outermost_contour")
    factory["writers"]["outer"] = shapefile.Writer(fname,
                                             shapeType = shapefile.POLYGON)
    
    fname = path.join(factory["SHPC"], "max_speed_contour")
    factory["writers"]["max_speed"] = shapefile.Writer(fname,
                                                 shapeType = shapefile.POLYGON)
    
    define_fields(factory["writers"])
    fname = path.join(factory["SHPC"], "ishape_last.txt")
    factory["writers"]["ishape_last"] = open(fname, "w")
    

# Loop on dates:
while True:
    if os.access(adt_file, os.R_OK):
        print("Processing", my_date, "...")
        os.symlink(adt_file, "adt.mat")
        subprocess.run(["matlab", "-batch", "inst_eddies"], check = True)
        matlab_data = sio.loadmat("adt_v6.mat", squeeze_me = True,
                                  variable_names = ("date_num", "X", "Y",
                                                    "Anticyclonic_Cell",
                                                    "Cyclonic_Cell", "ADT"))
        # matlab_data is a dictionary.
        os.remove("adt_v6.mat")
        assert datetime.date.fromordinal(matlab_data["date_num"] - 366) \
            == my_date

        for factory in factories:
            write(factory["writers"], matlab_data[factory["cell_name"]],
                  factory["cyclone"])

        os.remove("adt.mat")
    else:
        print("Missing file:", adt_file)
        
        for factory in factories:
            write(factory["writers"], [], factory["cyclone"])
        
    my_date += datetime.timedelta(1)
    if my_date > final_date: break
    d += 1
    basename = my_date.strftime(date_format)
    adt_file = path.join(dirname, basename)
    
nml = f90nml.Namelist()
nml["grid_nml"] = dict(corner_deg = [matlab_data["X"][0], matlab_data["Y"][0]],
                       nlon = matlab_data["X"].size,
                       nlat = matlab_data["Y"].size,
                       step_deg = [matlab_data["X"][1] - matlab_data["X"][0],
                                   matlab_data["Y"][1] - matlab_data["Y"][0]])

for factory in factories:
    fname = path.join(factory["SHPC"], "grid_nml.txt")
    nml.write(fname, force = True)
    fname = path.join(factory["SHPC"], "orientation.txt")
    
    with open(fname, "w") as f:
        if factory["cyclone"]:
            f.write("cyclones\n")
        else:
            f.write("anticyclones\n")
            
    for k, v in factory["writers"].items(): v.close()
